# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/pubsub.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import common_pb2 as src_dot_ray_dot_protobuf_dot_common__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='src/ray/protobuf/pubsub.proto',
  package='ray.rpc',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1dsrc/ray/protobuf/pubsub.proto\x12\x07ray.rpc\x1a\x1dsrc/ray/protobuf/common.proto\"\xbf\x02\n\nPubMessage\x12\x37\n\x0c\x63hannel_type\x18\x01 \x01(\x0e\x32\x14.ray.rpc.ChannelTypeR\x0b\x63hannelType\x12\x15\n\x06key_id\x18\x02 \x01(\x0cR\x05keyId\x12k\n\x1eworker_object_eviction_message\x18\x03 \x01(\x0b\x32$.ray.rpc.WorkerObjectEvictionMessageH\x00R\x1bworkerObjectEvictionMessage\x12_\n\x1aworker_ref_removed_message\x18\x04 \x01(\x0b\x32 .ray.rpc.WorkerRefRemovedMessageH\x00R\x17workerRefRemovedMessageB\x13\n\x11pub_message_oneof\":\n\x1bWorkerObjectEvictionMessage\x12\x1b\n\tobject_id\x18\x01 \x01(\x0cR\x08objectId\"]\n\x17WorkerRefRemovedMessage\x12\x42\n\rborrowed_refs\x18\x01 \x03(\x0b\x32\x1d.ray.rpc.ObjectReferenceCountR\x0c\x62orrowedRefs*I\n\x0b\x43hannelType\x12\x1a\n\x16WORKER_OBJECT_EVICTION\x10\x00\x12\x1e\n\x1aWORKER_REF_REMOVED_CHANNEL\x10\x01\x62\x06proto3'
  ,
  dependencies=[src_dot_ray_dot_protobuf_dot_common__pb2.DESCRIPTOR,])

_CHANNELTYPE = _descriptor.EnumDescriptor(
  name='ChannelType',
  full_name='ray.rpc.ChannelType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='WORKER_OBJECT_EVICTION', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WORKER_REF_REMOVED_CHANNEL', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=550,
  serialized_end=623,
)
_sym_db.RegisterEnumDescriptor(_CHANNELTYPE)

ChannelType = enum_type_wrapper.EnumTypeWrapper(_CHANNELTYPE)
WORKER_OBJECT_EVICTION = 0
WORKER_REF_REMOVED_CHANNEL = 1



_PUBMESSAGE = _descriptor.Descriptor(
  name='PubMessage',
  full_name='ray.rpc.PubMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='channel_type', full_name='ray.rpc.PubMessage.channel_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='channelType', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key_id', full_name='ray.rpc.PubMessage.key_id', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='keyId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='worker_object_eviction_message', full_name='ray.rpc.PubMessage.worker_object_eviction_message', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='workerObjectEvictionMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='worker_ref_removed_message', full_name='ray.rpc.PubMessage.worker_ref_removed_message', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='workerRefRemovedMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='pub_message_oneof', full_name='ray.rpc.PubMessage.pub_message_oneof',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=74,
  serialized_end=393,
)


_WORKEROBJECTEVICTIONMESSAGE = _descriptor.Descriptor(
  name='WorkerObjectEvictionMessage',
  full_name='ray.rpc.WorkerObjectEvictionMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='object_id', full_name='ray.rpc.WorkerObjectEvictionMessage.object_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='objectId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=395,
  serialized_end=453,
)


_WORKERREFREMOVEDMESSAGE = _descriptor.Descriptor(
  name='WorkerRefRemovedMessage',
  full_name='ray.rpc.WorkerRefRemovedMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='borrowed_refs', full_name='ray.rpc.WorkerRefRemovedMessage.borrowed_refs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='borrowedRefs', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=455,
  serialized_end=548,
)

_PUBMESSAGE.fields_by_name['channel_type'].enum_type = _CHANNELTYPE
_PUBMESSAGE.fields_by_name['worker_object_eviction_message'].message_type = _WORKEROBJECTEVICTIONMESSAGE
_PUBMESSAGE.fields_by_name['worker_ref_removed_message'].message_type = _WORKERREFREMOVEDMESSAGE
_PUBMESSAGE.oneofs_by_name['pub_message_oneof'].fields.append(
  _PUBMESSAGE.fields_by_name['worker_object_eviction_message'])
_PUBMESSAGE.fields_by_name['worker_object_eviction_message'].containing_oneof = _PUBMESSAGE.oneofs_by_name['pub_message_oneof']
_PUBMESSAGE.oneofs_by_name['pub_message_oneof'].fields.append(
  _PUBMESSAGE.fields_by_name['worker_ref_removed_message'])
_PUBMESSAGE.fields_by_name['worker_ref_removed_message'].containing_oneof = _PUBMESSAGE.oneofs_by_name['pub_message_oneof']
_WORKERREFREMOVEDMESSAGE.fields_by_name['borrowed_refs'].message_type = src_dot_ray_dot_protobuf_dot_common__pb2._OBJECTREFERENCECOUNT
DESCRIPTOR.message_types_by_name['PubMessage'] = _PUBMESSAGE
DESCRIPTOR.message_types_by_name['WorkerObjectEvictionMessage'] = _WORKEROBJECTEVICTIONMESSAGE
DESCRIPTOR.message_types_by_name['WorkerRefRemovedMessage'] = _WORKERREFREMOVEDMESSAGE
DESCRIPTOR.enum_types_by_name['ChannelType'] = _CHANNELTYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PubMessage = _reflection.GeneratedProtocolMessageType('PubMessage', (_message.Message,), {
  'DESCRIPTOR' : _PUBMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PubMessage)
  })
_sym_db.RegisterMessage(PubMessage)

WorkerObjectEvictionMessage = _reflection.GeneratedProtocolMessageType('WorkerObjectEvictionMessage', (_message.Message,), {
  'DESCRIPTOR' : _WORKEROBJECTEVICTIONMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerObjectEvictionMessage)
  })
_sym_db.RegisterMessage(WorkerObjectEvictionMessage)

WorkerRefRemovedMessage = _reflection.GeneratedProtocolMessageType('WorkerRefRemovedMessage', (_message.Message,), {
  'DESCRIPTOR' : _WORKERREFREMOVEDMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerRefRemovedMessage)
  })
_sym_db.RegisterMessage(WorkerRefRemovedMessage)


# @@protoc_insertion_point(module_scope)
